#include <string.h>
#ifdef PLATFORM_WINDOWS
	#include <windows.h>
	#include <gl\gl.h>
#else
	#include <GL/gl.h>
#endif

#include "Shared.h"

using namespace Obelisk;



#ifdef PLATFORM_WINDOWS
	extern const Obelisk::TCodeUnit *Obelisk::gEmptyString = L"";
#else
	extern const Obelisk::TCodeUnit *Obelisk::gEmptyString = "";
#endif



/**
Returns the number of code units in the code point at the specified offset in the string.
@param aData The string data.
@param aSize The size of the string data.
@param aOffset The offset into the string data of the code point.
@return int The number of code units in the code point.
*/
int CString::GetCodePointSize(const TCodeUnit* aData, int aSize, int aOffset)
{
	Assert(aSize >= 0 && aOffset >= 0 && aOffset < aSize)
	#ifdef PLATFORM_WINDOWS
		Assert(!(aData[aOffset] >= 0xdc00 && aData[aOffset] < 0xe000))
		if (!(aData[aOffset] >= 0xd800 && aData[aOffset] < 0xdc00))
			return 1;
		else
		{
			Assert(aOffset + 1 < aSize && aData[aOffset + 1] >= 0xdc00 &&
					aData[aOffset + 1] < 0xe000)
			return 2;
		}
	#else
		if ((aData[aOffset] & 0x80) == 0)
			return 1;
		else if ((aData[aOffset] & 0xe0) == 0xc0)
		{
			Assert(aOffset + 1 < aSize && (aData[aOffset + 1] & 0xc0) == 0x80)
			return 2;
		}
		else if ((aData[aOffset] & 0xf0) == 0xe0)
		{
			Assert(aOffset + 2 < aSize && (aData[aOffset + 1] & 0xc0) == 0x80 &&
					(aData[aOffset + 2] & 0xc0) == 0x80)
			return 3;
		}
		else if ((aData[aOffset] & 0xf8) == 0xf0)
		{
			Assert(aOffset + 3 < aSize && (aData[aOffset + 1] & 0xc0) == 0x80 &&
					(aData[aOffset + 2] & 0xc0) == 0x80 && (aData[aOffset + 3] & 0xc0) == 0x80)
			return 4;
		}
		else
			throw CException("Invalid string data");
	#endif
}



/**
Returns the number of code units in the code point at the specified offset in the string.
@param aData The string data.
@param aSize The size of the string data.
@param aOffset The offset into the string data of the last code unit in the code point.
@return int The number of code units in the code point.
*/
int CString::GetPrevCodePointSize(const TCodeUnit* aData, int aSize, int aOffset)
{
	Assert(aSize >= 0 && aOffset >= 0 && aOffset < aSize)
	#ifdef PLATFORM_WINDOWS
		Assert(!(aData[aOffset] >= 0xd800 && aData[aOffset] < 0xdb00))
		if (!(aData[aOffset] >= 0xdc00 && aData[aOffset] < 0xe000))
			return 1;
		else
		{
			Assert(aOffset - 1 >= 0 && aData[aOffset - 1] >= 0xd800 &&
					aData[aOffset - 1] < 0xdb00)
			return 2;
		}
	#else
		for (int i = 0; true; i++)
		{
			if ((aData[aOffset - i] & 0xc0) != 0x80)
			{
				switch (i)
				{
				case 0:
					Assert((aData[aOffset - i] & 0x80) == 0)
					return 1;
				case 1:
					Assert((aData[aOffset - i] & 0xe0) == 0xc0)
					return 2;
				case 2:
					Assert((aData[aOffset - i] & 0xf0) == 0xe0)
					return 3;
				case 3:
					Assert((aData[aOffset - i] & 0xf8) == 0xf0)
					return 4;
				}
				throw CException("Invalid string data");
			}
		}
	#endif
}



CString::CString()
{
	mReferenceCount = NULL;
	mData = NULL;
	mSize = 0;
}



CString::CString(int aSize, const TCodeUnit* aData)
{
	Assert(aSize == 0 || aData != NULL);
	if (aSize == 0)
	{
		mReferenceCount = NULL;
		mData = NULL;
		mSize = 0;
		return;
	}

	mReferenceCount = new int;
	*mReferenceCount = 1;
	mSize = aSize;
	mData = new TCodeUnit [mSize + 1];
	memcpy(mData, aData, mSize * sizeof(TCodeUnit));
	mData[mSize] = 0;

	#ifdef DEBUG
		#ifdef PLATFORM_WINDOWS
			for (int i = 0; i < mSize; i++)
			{
				Assert(!(mData[i] >= 0xdc00 && mData[i] < 0xe000))
				if (mData[i] >= 0xd800 && mData[i] < 0xdc00)
				{
					i++;
					Assert(i < mSize && mData[i] >= 0xdc00 && mData[i] < 0xe000)
				}
			}
		#else
			for (int i = 0; i < mSize; i++)
			{
				if ((mData[i] & 0xe0) == 0xc0)
				{
					Assert(i + 1 < mSize && (mData[i + 1] & 0xc0) == 0x80)
					int CodePoint = ((mData[i] & 0x1f) << 6) | (mData[i + 1] & 0x3f);
					Assert(CodePoint >= 0x80 && CodePoint < 0x800)
					i++;
				}
				else if ((mData[i] & 0xf0) == 0xe0)
				{
					Assert(i + 2 < mSize && (mData[i + 1] & 0xc0) == 0x80 &&
							(mData[i + 2] & 0xc0) == 0x80)
					int CodePoint = ((mData[i] & 0xf) << 12) | ((mData[i + 1] & 0x3f) << 6) |
							(mData[i + 2] & 0x3f);
					Assert(CodePoint >= 0x800 && CodePoint < 0x10000 &&
							!(CodePoint >= 0xd800 && CodePoint < 0xe000))
					i += 2;
				}
				else if ((mData[i] & 0xf8) == 0xf0)
				{
					Assert(i + 3 < mSize && (mData[i + 1] & 0xc0) == 0x80 &&
							(mData[i + 2] & 0xc0) == 0x80 && (mData[i + 3] & 0xc0) == 0x80)
					int CodePoint = ((mData[i] & 0x7) << 18) | ((mData[i + 1] & 0x3f) << 12) |
							((mData[i + 2] & 0x3f) << 6) | (mData[i + 3] & 0x3f);
					Assert(CodePoint >= 0x10000 && CodePoint < 0x110000)
					i += 3;
				}
				else
					Assert((mData[i] & 0x80) == 0)
			}
		#endif
	#endif
}



CString::CString(CString aString, int aCount)
{
	Assert(aCount >= 0)
	if (aString.IsEmpty() || aCount == 0)
	{
		mReferenceCount = NULL;
		mData = NULL;
		mSize = 0;
		return;
	}
	if (aCount == 1)
	{
		mReferenceCount = aString.mReferenceCount;
		(*mReferenceCount)++;
		mData = aString.mData;
		mSize = aString.mSize;
		return;
	}

	mReferenceCount = new int;
	*mReferenceCount = 1;
	mSize = aString.mSize * aCount;
	mData = new TCodeUnit [mSize + 1];
	for (int i = 0; i < aCount; i++)
		memcpy(mData + aString.mSize * i, aString.mData, aString.mSize * sizeof(TCodeUnit));
}



CString::CString(const CString& aString)
{
	mReferenceCount = aString.mReferenceCount;
	if (mReferenceCount != NULL)
		(*mReferenceCount)++;
	mData = aString.mData;
	mSize = aString.mSize;
}



CString::~CString()
{
	Release();
}



CString& CString::operator =(const CString& aString)
{
	Release();
	mReferenceCount = aString.mReferenceCount;
	if (mReferenceCount != NULL)
		(*mReferenceCount)++;
	mData = aString.mData;
	mSize = aString.mSize;
	return *this;
}



CString CString::operator +(const CString& aString) const
{
	if (aString.mSize == 0)
		return *this;
	if (mSize == 0)
		return aString;

	CString r;
	r.mReferenceCount = new int;
	*r.mReferenceCount = 1;
	r.mSize = mSize + aString.mSize;
	r.mData = new TCodeUnit [r.mSize + 1];
	memcpy(r.mData, mData, mSize * sizeof(TCodeUnit));
	memcpy(r.mData + mSize, aString.mData, aString.mSize * sizeof(TCodeUnit));
	r.mData[r.mSize] = 0;
	return r;
}



CString& CString::operator +=(const CString& aString)
{
	if (mSize == 0)
	{
		Release();
		mReferenceCount = aString.mReferenceCount;
		if (mReferenceCount)
			(*mReferenceCount)++;
		mData = aString.mData;
		mSize = aString.mSize;
	}
	else if (aString.mSize != 0)
	{
		int Size = mSize + aString.mSize;
		TCodeUnit *Data = new TCodeUnit [Size + 1];
		memcpy(Data, mData, mSize * sizeof(TCodeUnit));
		memcpy(Data + mSize, aString.mData, aString.mSize * sizeof(TCodeUnit));
		Data[Size] = 0;

		Release();
		mReferenceCount = new int;
		*mReferenceCount = 1;
		mData = Data;
		mSize = Size;
	}
	return *this;
}



bool CString::IsCharacter()
{
	return mSize != 0 && GetCodePointSize(mData, mSize, 0) == mSize;
}



bool CString::operator ==(const CString& aString) const
{
	if (mSize != aString.mSize)
		return false;
	return mSize == 0 ? true : memcmp(mData, aString.mData, mSize * sizeof(TCodeUnit)) == 0;
}



bool CString::operator !=(const CString& aString) const
{
	if (mSize != aString.mSize)
		return true;
	return mSize == 0 ? false : memcmp(mData, aString.mData, mSize * sizeof(TCodeUnit)) != 0;
}



int CString::Iterator(int aIndex) const
{
	int p = 0;
	for (int i = 0; i < aIndex; i++)
		p += GetCodePointSize(mData, mSize, p);
	return p;
}



void CString::Next(int& p) const
{
	p += GetCodePointSize(mData, mSize, p);
}



int CString::Last() const
{
	if (mSize == 0)
		return -1;
	return mSize - GetPrevCodePointSize(mData, mSize, mSize - 1);
}



void CString::Prev(int& p) const
{
	p -= GetPrevCodePointSize(mData, mSize, p - 1);
}



CString CString::GetChar(int p) const
{
	return CString(GetCodePointSize(mData, mSize, p), mData + p);
}